dnl ---------- OTOP.m4, bindings for the OTOP subproject --------------------
dnl Kernel bindings for a generic C kernel.
dnl
Provide({OTOP},{
dnl
Require({ANSIC})dnl
Require({version})dnl
Require({magic})dnl
Provide({config},{include(config)})dnl
dnl
dnl
dnl --------------------------- Structures --------------------------------
dnl

struct API_Signature {
   int32 Magic_Number ;
   char  Object_Name[20] ;
   char  Version[8] ;
};

struc API_Header {
   int32 TotalLength ;		/* total length of the file */
   int32 EntryPoint ;		/* address of object to return */
   int32 Offset	;		/* offset where to load it */
   int32 HeaderSize ;		/* size of header fields in file */
   int32 ExpandedSize ;		/* size when expanded to memory */
   int32 ___ ;			/* fill in */
   int32 CheckSum ;		/* checksum */
   int32 LongCheckSum ;		/* long checksum */
})


dnl
dnl
dnl ---------------------- C as an assembler -------------------------------


define({CELL},	CELLSIZE)		; a cell is 4 bytes long


dnl
dnl ---- abstraction of the i386 instruction set ----
dnl
define({Add},		{($1) += ($2) ;})dnl
define({Sub},		{($1) -= ($2) ;})dnl
define({__Mov__},	{($1) = ($2) ;})dnl
define({__Movb__},	{($1) = ($2) ;})dnl
define({__Movw__},	{($1) = ($2) ;})dnl
define({Neg},		{($1) = - ($1) ;})dnl
define({Xchg},		{R0 = $1 ; $1 = $2 ; $2 = R0 ;})dnl
define({},		{})dnl
define({},		{})dnl
dnl
dnl
define({point},	{ifelse({$2},{},{*((cell*)$1)},{((cell*)$1)[$2]})})dnl
dnl
dnl
define({PC},	{esi})dnl	program counter
define({DSP},	{edi})dnl	data stack pointer
define({DSINC},	{CELL})dnl	increment for data stack -- grow up
define({TOS},	{ebx})dnl	buffered top of stack
define({RSP},	{esp})dnl	return stack pointer
define({RSINC},	{-CELL)})dnl	increment for return stack -- grow down

define({FRAME},		{ebp})	; a frame pointer for heap manipulation
define({R0},		{eax})	; a scratch register
define({R0L},		{al})	; its lowest byte
define({R0H},		{ah})	; its second byte
define({R1},		{edx})	; second scratch register
define({COUNT},		{ecx})	; count register
define({COUNTLO},	{cl})	; count register -- low byte

define({INSTRTBL},	InstrT)		; table for bytecode threading

define({PopData},	{Mov({$1},TOS)
			 Add(DSP,-DSINC)})
define({PushData},	{Add(DSP,DSINC)
			 Mov(point(DSP),TOS)
			 Mov(TOS,{$1})})
define({PopReturn},	{ifelse({$1},,{Add(RSP,-RSINC)},{pop $1})})
define({PushReturn},	{push $1})
define({},		{})
define({},		{})
define({},		{})
define({},		{})

define({Enter},	{Add(PC,CELL)})
define({Next},	{jmp [PC]})


;;;;;;;;;;;;;;;;;;;;;;;;;; Defining FORTH words ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; defining a counted string
define({DefCString},
{__stringseg__({
locals({strlen},{strbeg},{strend},{
strbeg:
	.byte 	strlen
	defb($@)
strend:
strlen	=	strend-strbeg-1
})})})

;; defining a FORTH word


define({DefWord},{
ifelse({$2},,{$1},{$2}):
})





;;;;;;;;;;;;;;;;;;;;;;; Aerial (extra-light) threads ;;;;;;;;;;;;;;;;;;;;;;;;;
Struc({AT},{
; AerialThread	; These threads are lighter than light: Aerial !
   int32 SAVE_EBX	; initial Top of Stack
   int32 SAVE_ESI	; Data Stack Pointer
   int32 SAVE_EDI	; Return Stack Pointer
   int32 SAVE_EBP	; Those two are reserved for the
   int32 SAVE_ESP	;; threading mechanism.
   int32 SAVE_RESP	; ESP save when going ret-threaded
   int32 STATE		; flags used for thread sync'
   int32 MAX_ESI	; bottom of data stack
   int32 MAX_EDI	; bottom of return stack
   int32 LINKBACK	; Link Back to thread object
})



/*** Some instructions ***/

define({DefCompareWord0},
{DefWord({$1},{$2},{$3})
   Enter
   TOS=(TOS $4 0)?FALSE:TRUE ;
   Next
})dnl
define({DefCompareWord},
{DefWord({$1},{$2},{$3})
   Enter
   TOS=(NOS $4 TOS)?FALSE:TRUE ;
   Nip
   Next
})dnl
dnl
dnl
dnl
})dnl --- End of OTOP.m4 ---
dnl
