include(macro.m4)
Require({i386})

; BIOS-based Loader for the 32-bit TUNES initial object.
; Copyright (c) Franois-Ren "Far" Bn Rideau Dang-Vu 1994,1995

; Please Read the info/BIOS-API.nfo file (section * loader.asm::).




;;; Debugging defines:

undefine({__NO_USE_INT_13H__})
undefine({DEBUG})		; debug messages
undefine({__DEBUG__})		; heavy debug
undefine({CLEARSCREEN})		; clear it before loading
define({MESSAGES})		; Do print messages; currently a must do



;;; Defines that may change.

; We mustn't really use these. The Boot Record need it, though.
LOADER_SEG_SIZE	 =  0x0F00	;(60KB) ; size in para for the loader segment.
LOADER_STACK_TOP =  LOADER_SEG_SIZE*16	; SP in the same segment
STACKSIZE	 =  8192		; stack size when using LILO
LOADER_SECTS	 =  3			; size of the loader in sectors.
LOADEROFS	 =  0x100		; load the loader as a .COM file.
LOADER_UDATA	 =  0			; no udata area, as 0x80-0x100 is used.
KERNEL_SECTS	 =  512			; Size of Kernel+API to load.
MIN_KERNEL_SPACE =  300			; KB needed by the kernel
MAX_KERNEL_SPACE =  16000		; KB asked by the kernel
MAX_TRY		 =  10		        ; number of retries before aborting I/O

APIOFS		= 0x1000
COMMAND_LINE	= 0x0800


;;; Defines that won't change
LOG_SECT_SIZE	=	9		; Size of sectors (low-level standard)
SECT_SIZE	=   1 << LOG_SECT_SIZE	; 512

MAP_FILE	 =  	LOADEROFS+SECT_SIZE*LOADER_SECTS
		; sector-aligned just after the loader: loaded
		; by the boot sector as virtual part of the loader

; LOAD MODEs:
; * 'D' for loading a DOS file as the API+Kernel. LoadArg points to the
;  name (11 bytes) of the system file containing the API and the Kernel
;  system means the sectors are contiguous and unmovable, as with usual
;  DOS files...
; * 'H' for using an overformatted floppy's hidden sectors as the API+Kernel.
;  LoadArg contains real max tracks (hi) and real max sects/tracks (lo).
; * 'C' for using contiguous directories after the current one. LoadArg is
;  The number of the first sector (relative to hidden_sectors).
; * 'M' for LILO-like mapping. The map file is already loaded at address
;  MAP_FILE
; * 'L' for just moving a LILO (or LOADLIN) loaded kernel. Why rewrite LILO ?


;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Code first ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
USE16

		.text
		use16
		.org	LOADEROFS	; .COM file format

.globl _main
_main:
		;br	Init	; Init just follows.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Loader DATA ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
.data


;.bss	;;I manually allocated data; I should have used macros...

newheap(0,0x80)
ConvMem		=	allot(0,4)	; Total Conventional memory in bytes
AvConvMem	=	allot(0,4)	; Available Conv. mem in bytes
XtdMem		=	allot(0,4)	; Total Extended memory in bytes
AvXtdMem	=	allot(0,4)	; Available Extended mem in bytes

BaseConvMem:	.long	0x1000  ;4096	; Page aligned, reserve 2K for BIOS.
BaseXtdMem:	.long	0x110000;1MB+64KB ; Page aligned, reserve low
					; memory & HMA for dos emulator.

CPUType 	=	allot(0,2)	; AH:CPU, AL:NPU (see mem386.inc)


;;; Inits data for Next_cluster
Init_Sector_Read =	allot(0,2)

;;; Returns the next block of sector to read in int 13h -able registers.
;;; al=0 if no more clusters.
Next_cluster	=	allot(0,2)


;;; we assume we could fill the user-defined buffer
;;; if not, then bufferize those routines with an overfillable
;;; buffer (overfilling: up to the maximal value given by Next_cluster).
over_fill	=	alallot(0,4)	; 32-bit linear address !!!
over_fill_len	=	alallot(0,4)	; 32-bit length


;;; We also assume no more than 64 KB overfilled buffer, no 64 KB boundary
;;; crossed, no pointer wraps, new value of buffer not interfering with old
;;; overfull buffer. (that's all :)

ErrorCode	=	allot(0,1)	; must be all on the same disk.


;;; Load Mode 'H' ; for fdformatted floppies only
;;; (can anyone achieve over-formatted hd's ?)
Max_Tracks_on	=	allot(0,1)	; max. number of "official" tracks
CurrentTrack	=	allot(0,2)	; hi: track, lo: head


;;; Load Mode 'M'
pCluster	=	allot(0,2)


;;; Load Mode 'C'
CurrentSector	=	alallot(0,4)
MaxSector	=	alallot(0,4)

;;; Debugging LOADER
Int13Handler	=	allot(0,4)

;;; Load Mode 'D'
DataArea	=	allot(0,2)
CurrentCluster	=	allot(0,2)
SectorInCluster	=	allot(0,2)
FAT_Reader	=	allot(0,2)	; routine to read 12 or 16 bit FAT
FAT_Mapped_Beg	=	allot(0,2)	; for 16 bit FATs, which part is mapped
FAT_Mapped_Len	=	allot(0,2)


;;; Load Mode 'L'
; Nothing more is needed ;


;;; API Header information	; see APIHeader.asm

API_Signature	=	0x1000	; Signature_Struc  es:0x1000
API_Header	=	0x1020	; API_Header_Struc es:0x1020
API_Magic	=	API_Signature+AS_Magic_Number	; dword
API_Length	=	API_Header+AH_TotalLength	; dword
API_EntryPoint	=	API_Header+AH_EntryPoint	; dword


;;; LILO style loading
INITSEG		=	0x9000
SYSSEG		=	0x1000
SETUPSEG	=	0x9020
SETUPOFS	=	0x0000
ADJSETUPSEG	=	SETUPSEG-((LOADEROFS-SETUPOFS)>>4)
SYSSIZE		=	0x7F00
CL_MAGIC_ADDR	=	0x0020	; in INITSEG
CL_MAGIC	=	0xA33F
CL_OFFSET	=	0x0022
Lxsig_1		=	0xAA55
Lxsig_2		=	0x5A5A

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; ARGUMENTS ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; see bootrec.asm for full details...

;;; DOS info table

bytes_per_sect	=	0x0B	; word
sectors_per_clu	=	0x0D	; byte
reserved_sect	=	0x0E	; word
num_FATs	=	0x10	; byte
entries_rootdir	=	0x11	; word
total_num_sect	=	0x13	; word
media_descr	=	0x15	; byte
sect_per_FAT	=	0x16	; word
sect_per_track	=	0x18	; word
num_heads	=	0x1A	; word (actually, byte)
hidden_sect	=	0x1C	; dword
d_total_n_sect	=	0x20	; dword

;;; TUNES Loading Parameters

LoadMode	=	0x24	; byte
LoadArg		=	0x25	; word
BIOS_DL		=	0x27	; byte
Clusters	=	0x28	; dword

;;; Kernel Header

Kernel_Magic	=	0x00	; dword es:0x0000
Kernel_Length	=	0x20	; dword es:0x0020


;;;;;;;;;;;;;;;;;;;;;;;;; Initialization Routines ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
.text

;;;;; These procedures are executed only once, so I macro'ed them.

define({AdjustCS},{
  ;;; Be sure that we got loaded at org LOADEROFS
	call ACS
     ACS:
	pop dx			; get return address (must be ACS)
	mov cx,#ACS
	sub cx,dx		; compute the offset we want.
	jz GoodCS
	STOP(fs,0,38,15,'B,0x0F)
	mov si,cx
	rept(4,{
		sar cx,#1
	})
	mov bx,cs
	sub bx,cx
	mov ds,bx		; adjust data segment
	and si,0x000F
	jnz InitConfused
	push bx
	mov ax,#LILO_test
	push ax
	STOP(fs,0,39,15,'L,0x0F)
	retf

InitConfused:
	STOP(fs,0,38,14,'C,0x0A)
	STOP(fs,0,39,14,'o,0x0A)
	add si,#defstring({"I dunno where I am !"},CRLF,0)
	STOP(fs,0,40,14,'n,0x0A)
	STOP(fs,0,41,14,'f,0x0A)
Exit2:
	br Exit

GoodCS:
	STOP(fs,0,38,16,'G,0x0F)
})


define({Init_LOADER},{
  ; Test CPU & FPU type
  LibInline(TestCPU)
  mov [CPUType],ax
  __IF__(b)
    cmp ah,#3
  __THEN__
    lsa(si,{"This TUNES version requires a 32 bit (386+) processor"},0)
    br Exit
  __ENDIF__
 ifdef({__NO_USE_INT_13H__},{
    push es
    push word #0
    pop es
    seg es
    mov eax,[0x13*4]
    mov [Int13Handler],eax
    pop es
 })
})


define({Parse_Options},{
  ;;; Parse_Options
  ; Verify that DOS aint loaded
   STOP(fs,0,40,13,'O,0xF)
   __IF__(nz)
	cmp byte [0x80],#0x99	; The bootrec ought to have put this value
				; 0x99>=0x80, thus COMMAND.COM won't put it. '
   __THEN__
      lsa(si,{"From DOS, use LOADLIN !"},CRLF,0)
      br Puts
   __ENDIF__
   mov al,[LoadMode]
   mov bx,#D_Init_disk
   mov dx,#D_Next_cluster
   cmp al,#'D
   jz Parsed
   mov bx,#M_Init_disk
   mov dx,#M_Next_cluster
   cmp al,#'M
   jz Parsed
   mov bx,#C_Init_disk
   mov dx,#C_Next_cluster
   cmp al,#'C
   jz Parsed
   mov bx,#H_Init_disk
   mov dx,#H_Next_cluster
   cmp al,#'H
   jz Parsed
   mov bx,#L_Init_disk
   mov dx,#L_Next_cluster
   cmp al,#'L
   jz Parsed
 Load_Mode_Unsupported:
   __dataseg__
     Mode_not_supported:	.ascii "Load Mode "
     Unsupported_Mode:		.ascii "X not supported."
     db_zero:			.byte 0
   __codeseg__
   mov [Unsupported_Mode],al
   mov si,#Mode_not_supported
   br Exit
 Parsed:
   mov [Init_Sector_Read],bx
   mov [Next_cluster],dx
})

define({Init_Message},{
  ;;; Init_Message
	__dataseg__
	  sLoading:	.ascii "Loading using Mode "
	  Loading_Mode:	defb('X,"..",0)
	__codeseg__
	mov al,[LoadMode]
	mov [Loading_Mode],al
	mov si,#sLoading
	PUTS
})

define({Get_Memory},{
   STOP(fs,0,41,13,'S,0xF)
  ;;; Get Total extended memory
	mov ah,#0x88
	int #0x15	; result in KB
	movzx eax,ax
	shl eax,#10
	mov [XtdMem],eax; result in bytes
  ;;; Compute available x. mem.
	sub eax,#65536	; Don't put anything in the HMA zone '
	mov [AvXtdMem],eax
  ;;; Get Total conventional memory
	int #0x12	; result in KB
	movzx eax,ax
	shl eax,#10
	mov [ConvMem],eax
  ;;; Compute Available conventional memory
    ; assuming the boot record loaded us topmost is *not*
    ; possible, since LILO could have done otherwise
	shr eax,#4	; convert to para
	sub ax,#0x1000	; 64KB in para
	push ds
	mov ds,ax
	mov ebx,#0x10000
	cmp dword [bx-4],#MAGIC_MARKER
	jnz AvConvMemFound
	sub ebx,6
     Run_Res_ConvMem:
	mov cx,[bx]
	jcxz AvConvMemFound
	sub bx,cx
	jae Run_Res_ConvMem
	mov ebx,#0x10000
     AvConvMemFound:
	shl eax,#4
	add eax,ebx
	pop ds
	mov [AvConvMem],eax
})

define({Init_Disk_Read},{
   ;;; Init sector count
	call [Init_Sector_Read]
})

define({LOAD_API},{
   ; Note: for the moment, the API should not be larger than 60 KB,
   ;  and even 30 KB to avoid a 64KB DMA limit with sector prefetching ;
   ;  Else a BIOS wrapper will have to buffer sector reading somewhere
   ;  (the 2KB between 0x00800 and 0x01000 ?) when some wrapping is detected.

   STOP(fs,0,42,13,'E,0xF)
   ;;; Init The buffer
	clr(ebp)
	mov es,bp	; 4KB
	mov ebx,#APIOFS
   ;;; Load the Header
	mov bp,#64	; Header length ((!) e- of ebp=0 already)
	call DoReadOnce
   ;;; Get the length from the header
	zLZonez({"Header:",CRLF},es,bx,16,{CRLF})
	__IF__(nz)
	  seg es
	  cmp dword [bx],#MAGIC_PC_API ; API_Magic
	__THEN__
	  zdwoz({"Magic number "},es:[bx],{CRLF})
	  ;z_z
	  lsa(si,{"API file fails magic number test."},CRLF,0)
	  jmp Exit3
	__ENDIF__
	seg es
	mov ebp,[API_Length]
	__IF__(a)
	  clr(edx)
	  mov dx,ds
	  shl edx,#4
	  sub edx,ebx
	  cmp ebp,edx ; we cannot it if bigger than free memory.
	__THEN__
	  lsa(si,{"API Too big"},CRLF,0)
	Exit3:
	  zzz({br},{jmp}) Exit
	__ENDIF__

   ;;; Load the entire API.
	STOP(fs,0,38,14,'A,0x0F)
	add ebx,#64
	sub ebp,#64
	call DoReadOnce
   ;;; ... ToDo:
   ;;; - add relocation (?)/ header parsing
   ;;; - add decompression (?)
})

define({LUKI_GO_API},{
   ;;; push API parameters
	push dword [XtdMem]
	push dword [ConvMem]
	push dword [AvConvMem]
	push word [CPUType]
   ;;; Jump into the API.
	PUTS(".",CRLF)
	clr(bx)
	push bx
	mov es,bx
	mov ds,bx
	seg es
	push word [API_EntryPoint]
	; (!) we should test that it is inferior to the API size ;
	STOP(fs,0,38,10,'J,0x0F)
	STOP(fs,0,39,10,'u,0x0F)
	STOP(fs,0,40,10,'m,0x0F)
	STOP(fs,0,41,10,'p,0x0F)
	zzz({
	  ;z_z
	  push ds
	  pusha
	  push cs
	  pop ds
	  seg es
	  mov di,[API_EntryPoint]
	  zZonez({"API Entry:"},es,di,16,{})
	  CGETC({CRLF,{"Hit any key to go TUNES !"},CRLF})
	  popa
	  pop ds
	})
	retf
})

define({Do_Exit},{
	call PrCRLF
	PUTS
	CGETC({CRLF,{"Hit any key to reboot"}})
	int #0x19
})


;;; Actual code ;;;

Init:

ifdef({__DEBUG__},{
	push #0xB800
	pop fs
	STOP(fs,0,38,13,'M,0x0F)
})

ifdef({CLEARSCREEN},{
	mov	ax,#0x0003	; 80x25, 16 colors
	int	#0x10
})

 ;;; both initialize stack and command line 
	clr(ax)
	mov ss,ax		; adjust stack segment
	mov sp,#COMMAND_LINE+2
	push ax

	AdjustCS

 ;__({"Testing {z_z}"},CRLF)
	;z_z

 __({"Initing self"},CRLF)
	Init_LOADER
	zwoz({"CPU:"},[CPUType],{CRLF})

forget({
 __({"Testing Int 13h",CRLF})
	push word #0
	pop es
	zaddrz({" at "},{es:[0x13*4+2]},{es:[0x13*4]},{CRLF})
	clr(ah)
	clr(dl)
	Int13h
	z_z
})

 __("Looking at options",CRLF)
	Parse_Options

 __({"Hello, World !"},CRLF)
	Init_Message

 __({"Seeing how much memory is left"},CRLF)
	Get_Memory
	zdwoz({"Found 0x"},[ConvMem],{{" bytes of conventional memory,"},CRLF})
	zdwoz({"and 0x"},{[XtdMem]},{{" bytes of extended memory"},CRLF})

 __({"Initialize disk reading"},CRLF)
	Init_Disk_Read

 __({"Load the API at 4KB"},CRLF)
	LOAD_API
	zdwoz({"Done loading a 0x"},es:[API_Length],{" long API",CRLF})

 __({"Jumping into the API"},CRLF)
	LUKI_GO_API

Exit:
 __({"Aborting..."},CRLF)
	Do_Exit			; On error, exit



define({Test_LILO},{
	STOP(fs,0,39,13,'O,0x0F)
	__IF__(nz)
	  cmp bx,#ADJSETUPSEG
	__THEN__
	  br GoodCS
	__ENDIF__

	End_LILOs_work
})



define({End_LILOs_work},{
;;; End_LILOs_work:		; stolen from linux/boot/setup.S


   ;;; Init registers for LILO
 	STOP(fs,0,38,13,'L,0x09)
 	STOP(fs,0,39,13,'I,0x09)
 	STOP(fs,0,40,13,'L,0x09)
 	STOP(fs,0,41,13,'O,0x09)


   ;;; Load what remains of the loader (as under linux)
   ;;; bx will contain the beginning of the system loaded by LILO

	STOP(fs,0,1,5,'S,0x9)
	mov	bx,#SYSSEG
	mov	ax,#INITSEG
	mov	es,ax

   ;;; Check signature at end of setup
   ;;; if present, nothing to do.
	cmp	word [setup_sig1],#Lxsig_1
	jne	bad_sig
	cmp	word [setup_sig2],#Lxsig_2
	jne	bad_sig
	jmp	good_sig


bad_sig:
; Move rest of setup code/data to here
	seg es
	mov dh,[497]			; get setup sects from boot sector
	sub dh,#4			; LILO loads 4 sectors of setup
	xor dl,dl			; dh is #sector; dx is #words
					; (!) assumes sectors are 512 bytes
	mov cx,dx
	shr dx,#3		; convert to segment
	add bx,dx

	;z_z
	;STOP(fs,0,3,5,'M,0x09)
	pushm(ds,es)
	pushm(#SYSSEG,#SETUPSEG)
	popm(ds,es)
	mov di,2048		; dest: after the 4 sect already loaded.
	clr(si)			; source: beginning of the kernel image
	rep
	movsw
	popm(ds,es)

good_sig:
	zLZonez({"at bx:",CRLF},bx,0,16,{CRLF})
	zLZonez({"at SYS:",CRLF},SYSSEG,0,16,{CRLF})

   ;;; Init parameters for LILO
 	STOP(fs,0,38,24,'L,0x09)
 	STOP(fs,0,39,24,'I,0x09)
 	STOP(fs,0,40,24,'L,0x09)
 	STOP(fs,0,41,24,'O,0x09)
	movzx eax,bx
	shl eax,#4
	mov dword [over_fill],eax
	clr(eax)
	seg es
	mov ax,[500]
	shl eax,#4
	mov dword [over_fill_len],eax
	mov byte [LoadMode],#'L
	mov byte [0x80],#0x99
	z__z({"Readjusted from LILO:",CRLF},SST)
	push ds
	__IF__(z)
	  push es
	  pop ds
	  cmp word [CL_MAGIC_ADDR],#CL_MAGIC
	__THEN__
	  STOP(fs,0,78,24,'C,0x0A,SST)
	  push ss	; (!) assumes ss is 0
	  pop es
	  mov si,[CL_OFFSET]
	  mov di,#COMMAND_LINE
	  mov cx,#APIOFS-COMMAND_LINE-1
	  rep
	  movsb
	  stosb	; ending 0 (in case of)
	__ENDIF__
	pop ds

   ;;; Done
	STOP(fs,0,41,20,'D,0x09)
	br GoodCS
})

LILO_test:
	Test_LILO


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Console I/O ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
__codeseg__

ifdef({MESSAGES},{

Putc	=	cPutc
Getc	=	cGetc


; Adapted from MicroSoft's own boot sector
; DS:SI=ASCIIZ string to print
; AX spoilt, SI points just after the string-ending zero, 7 put into BX.
Puts:
	;mov	al,[cs:si]
	;inc	si
	cld
	lodsb
	or	al,al
	jz	Return
	PUTC
	jmp	Puts

forget({
Beep:
; Echo a BEEP to the the Console
	mov	al,#7
;	jmp	cPutc
})

; Outputs a character to console
; spoils AX and DI, keeps si
cPutc:
	pushm(a,es)	; no pusha: may be used on 8088 ;;;
	mov	ah,#0xE	; function Teletype
	mov	bx,#7	; attribute
	int	#0x10	; Video BIOS call
	popm(a,es)
Return:
	ret

PrCRLF:
	PUTC(#13)
	mov al,#10
	jmp Putc


zzz({

Space:
	mov al,#' 
	jmp Putc

__dataseg__
  RegsMsg:	.ascii "DI   SI   BP   SP   "
		.ascii "BX   DX   CX   AX   "
		.ascii "FLGS CS   DS   ES   "
		.ascii "SS   IP   STK1 STK2"
		.byte CRLF,0
  ERegsMsg1:	.ascii "EDI      ESI      EBP      ESP      "
		.ascii "EBX      EDX      ECX      EAX      EFLAGS"
		.byte CRLF,0
  ERegsMsg2:	.ascii "CS   DS   ES   SS   IP   STK1 STK2 STK3 "
		.ascii "STK4 STK5 STK6 STK7 STK8 STK9 STKA STKB"
		.byte CRLF,0
__codeseg__

DumpERegs:
	push ss
	push es
	push ds
	push cs
	pushfd
	pushad
	push ss
	pop es
	push cs
	pop ds
	mov si,#ERegsMsg1
	PUTS
	mov di,sp
	mov cx,#9
	call PutLZone
	mov si,#ERegsMsg2
	PUTS
	mov cx,#16
	call PutWZone
	popad
	popfd
	pop es	;no pop cs above 8086
	pop ds
	pop es
	pop ss
 Return2:
	ret

DumpRegs:
	push ss
	push es
	push ds
	push cs
	pushf
	pusha
	push ss
	pop es
	push cs
	pop ds
	mov si,#RegsMsg
	PUTS
	mov di,sp
	mov cx,#16
	call PutWZone
	call PrCRLF
	popa
	popf
	pop es	;no pop cs above 8086
	pop ds
	pop es
	pop ss
	ret


PutLZone_:
	call Space
PutLZone:
	seg es
	mov eax,[di]
	add di,4
	call PutHexDword
	loop PutLZone_
	ret

PutWZone_:
	call Space
PutWZone:
	seg es
	mov ax,[di]
	inc di
	inc di
	call PutHexWord
	loop PutWZone_
	ret

PutZone_:
local({Ret},{{
	call	Space
PutZone:
	seg es
	mov al,[di]
	inc di
	call PutHexByte
	jcxz Ret
	call Space
	loop PutZone_
Ret:	ret
}})

PutAddr:
;; Prints the contents of es:di
	push di
	mov ax,es
	call PutHexWord
	PUTC(#':)
	pop ax
	jmp PutHexWord


PutHexDword:
	rol eax,#16
	call PutHexWord
	rol eax,#16
;	jmp PutHexWord


PutHexWord:
; Prints the contents of AX in hexa.
; everything but F saved
		xchg al,ah
		call PutHexByte
		xchg al,ah
;		jmp PutHexByte

})

; Prints the contents of AL in hexa.
; saves AX.
PutHexByte:
	push cx
	mov cl,#4
	rol al,cl
	call PutHexNybble
	rol al,cl
	pop cx
;	jmp PutHexNybble


; Prints to BIOS console the hexa nybble in low AL
; saves AX, puts 7 into BX
PutHexNybble:
	push ax
	and al,#0xF
	add al,#'0
	cmp al,#0xA+'0
	jb do_putnyb
	add al,#'A-'0-0xA
 do_putnyb:	
	PUTC
	pop ax
	ret


cGetc:
	clr(ah)
	int #0x16
	ret


forget({
sPutc:
	stosb
	ret

sGetc:
	lodsb
	ret

})


},{

Putc	=	Return
Puts	=	Return
PrCRLF	=	Return
cGetc	=	Return
Return:	ret

})	; MESSAGES

;;;;;;;;;;;;;;;;;;;;;;;;;; Misc. Subroutines ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

__codeseg__

define({_z_},{
   zzz({
	push ds
	push ss
	pop ds
	pusha
	PUTS($1)
	popa
	pop ds
	z_z({},{$2})
   })
})


edx_to_ds_si:	; spoils dx,esi
	mov esi,edx
	and si,#0x000F
	shr edx,#4
	mov ds,dx
	ret
edx_to_es_di:
	mov edi,edx
	and di,#0x000F
	shr edx,#4
	mov es,dx
	ret


safe_movsb_32:
; eax: source
; ebx: dest
; ecx: length in byte
; all regs conserved, but flags
	_z_({"safe_movsb:",CRLF})
	jcxz Return_0
     ;;; compare source & destination
	cmp ebx,eax
	jz Return_0
	pushm(es,ds,ad)
	mov ebp,#0x8000
	ja sm_dest_gt_source
;sm_source_gt_dest:
	_z_({"source gt dest:",CRLF})		; I couldn't write "s>d" ):
	cld
	__WHILE__(a)
	;;; setting up source/destination
	  mov edx,eax		; setting up ds:si from eax
	  call edx_to_ds_si
	  mov edx,ebx		; setting up es:di from ebx
	  call edx_to_es_di
	;;; more than 32KB ?
	  cmp ecx,ebp
	__DO__
	  push ecx
	  mov cx,#0x2000	; setting up cx
	  rep
	  movsd
	  pop ecx
	  add eax,ebp
	  add ebx,ebp
	  sub ecx,ebp
	__WEND__

finish_movsb:
	push cx
	__IF__(nz)
	  shr cx,#2
	__THEN__
	  rep
	  movsd
	__ENDIF__
	pop cx
	__IF__(nz)
	and cx,#3
	__THEN__
	  rep
	  movsb
	__ENDIF__
Return_1:
	popm(es,ds,ad)
Return_0:
	ret

; assuming dest > source
 sm_dest_gt_source:
	_z_({"source < dest:",CRLF})
	add eax,ecx ; moving downward
	add ebx,ecx
	std
	__WHILE__(a)
	;;; more than 32KB ?
	  cmp ecx,ebp
	__DO__
	;;; setting up source/destination
	  sub eax,ebp
	  sub ebx,ebp
	  sub ecx,ebp
	  mov edx,eax		; setting up ds:si from eax
	  call edx_to_ds_si
	  add si,bp
	  mov edx,ebx		; setting up es:di from ebx
	  call edx_to_es_di
	  add di,bp
	;;; moving
	  push ecx
	  mov cx,#0x2000	; setting up cx
	  rep
	  movsd
	  pop ecx
	__WEND__

	sub eax,ecx
	sub ebx,ecx
	mov edx,eax		; setting up ds:si from eax
	call edx_to_ds_si
	add si,cx
	mov edx,ebx		; setting up es:di from ebx
	call edx_to_es_di
	add di,cx
	jmp finish_movsb


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; DISK PROCESSING ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
ifdef({__NO_USE_INT_13H__},{
define({Int13h},{
    __({"Int 13h called:"},CRLF)
    z_z
    pushf
    push cs
    call Do_Int13h
    __({"Int 13h returned:"},CRLF)
    z_z
})
Do_Int13h:
    cli
    jmp [Int13Handler]
},
{
define({Int13h},{
    ;__({"Int 13h called:"},CRLF)
    ;z_z
    int #0x13
    ;__({"Int 13h returned:"},CRLF)
    ;z_z
})})


;;;;;;;;;;;;;;;;;;;;;;;;;;;; General Subroutines ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

SectorNumToBIOS_Args:
; In: eax (0 for first sector)
; Out: BIOS sector identifier (cl,dh,ch)

	;__({"SectorNumToBIOS_Args Input:"},CRLF)
	;z__z

	clr(edx)
	mov ebx,edx
	mov bl, [sect_per_track]
	div ebx		; quotient in eax, remainder in edx
	mov cl,dl	; remaining: sector
	inc cl		; they are from one on
	mov bl,[num_heads]
	clr(edx)
	div ebx
	mov dh,dl	; remaining: head
	mov ch,al
	shl ah,#6
	or cl,ah

	;__({"SectorNumToBIOS_Args Output:"},CRLF)
	;z__z
	STOP(fs,0,79,1,'S,0x0F,SST)
	ret


;BIOSToSectorNum:
; In: BIOS sector identifier
; Out: eax
;	dec cl			; sectors begin with 1
;	movzx ebx,cl		; ebx := sector number
;	and bl,#0x3F
;	shr cl,#6		; ax := cylinder number
;	mov al,ch
;	mov ah,cl
;	movzx ecx,dh		; ecx := head number
;	mul byte [num_heads]	; dx,ax := cylinder number * heads
;	push dx			; eax := dx,ax
;	push ax
;	pop eax
;	add eax,ecx
;	movzx edx,byte [sect_per_track]
;	mul edx
;	add eax,ebx
;	ret


Read_Sectors_With_Retries:
;; ax,bx,cd,dx,bp registers unchanged.
;; si (?), di spoilt
;; CF set if error (then see [ErrorCode])
 ;__({"Reading Sectors"},CRLF)
 ;z_z
   mov di,MAX_TRY	; maximum number of retries
   mov ah,#2	; Read
  Read_Try:
   push di
   push ax
   Int13h
   mov [ErrorCode],ah
   pop ax
   jnc Read_ok
   push ax
   clr(ah)		; reset controller
   push bp
   Int13h
   pop bp
   pop ax
   pop di
   dec di
   jnz Read_Try
 ;Too_Many_Retries:
   stc
   ;zbyz({"Disk I/O Error 0x"},[ErrorCode],{CRLF})
   ret
 Read_ok:
   pop di
   ret

Read_Sectors:
;;Input:
;; ebx is linear address of destination
;; al,cx,dx are Cluster number for BIOS int 13
;;Action:
;; read cluster,
;; managing the forbiddance of crossing 64KB DMA segment boundaries
;;Output:
;; all registers spoilt

   pushm(ds,es)
   movzx bp,al	  ; save BIOS parameter

 ;;; compute current DMA segment
   mov eax,ebx	  ; (!) < 1M
   shr eax,#4	  ; now e- part of eax is zero
   and ax,#0xF000 ; 64KB-aligned segment
   mov es,ax

 ;;; compute remaining length in it
   mov di,bx
   neg di		; length in segment before boundary

 ;;; can we do everything in one read ?
   shr di,#LOG_SECT_SIZE
   cmp bp,di
   jbe finish_RS

 ;;; no, read whatever can be in current DMA segment
   mov ax,di		; usuable length before segment
   sub bp,ax
   call Read_Sectors_With_Retries
   jc RS_xxx
   add cl,al

 ;;; if there is unaligned data, read it in a buffer and move it
   __IF__(nz)
     test bx,#SECT_SIZE-1
   __THEN__
     push es
     or bx,#~(SECT_SIZE-1)
     push bx
   ;;; read in a buffer
     push cs
     pop es
     mov bx,#BufferBeg
     mov al,#1
     call Read_Sectors_With_Retries
     pop ax
     pop di
     jc RS_xxx
   ;;; mov from the buffer to its rightful destination
     push cx
     push cs			; source
     pop ds
     mov si,#BufferBeg
     add ax,#0x0800		; dest
     mov es,ax
     sub di,#0x8000
     mov cx,#SECT_SIZE/4	; len
     rep			; move
     movsd
     pop cx
   ;;; adjust arguments for continuation
     mov bx,di
     inc cl
     dec bp
     clc
     jz RS_xxx
   __ELSE__
     mov ax,es
     add ax,#0x1000
     mov es,ax
     clr(bx)
   __ENDIF__

 ;;; a cluster can never be more than 32KB long anyway,
 ;;; so only one boundary may have been crossed...

finish_RS:
   mov ax,bp
   call Read_Sectors_With_Retries
RS_xxx:
   popm(ds,es)
   ret


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Global Routines ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Read_Once:
; Input:
; * (Min) Length to read in ebp
; * buffer at linear address ebx
; Output:
; * buffer (overwritten)
; * if buffer changed place, data moved without interference
; * ebp,ebx: preserved
; * ecx: amount of data actually read (overfull if > ax; end of file if < ax)
; * other registers spoilt
; *** NOTE: if you use data over bx+ax, you must set [over_fill_len] to
; ***  zero, or set it to another good value, and modify [over_fill]

 ;;; That's Structured programming in assembly ;
 __({"Read Once Input: "},CRLF)
 zdwoz({"asked 0x"},ebp,{" bytes at 0x"})
 zdwoz({},ebx,{CRLF})

 ;;; load old buffer address
   mov eax,dword [over_fill]
   mov ecx,dword [over_fill_len]
   zdwoz({"Already were 0x"},ecx,{" bytes at 0x"})
   zdwoz({},eax,{CRLF})
   call safe_movsb_32 ; moving to ebx
   __({"Buffer now ok"},CRLF)

 ;;; load more as needed
   push ebx
   __WHILE__(b)	; While buffer strictly not overfull enough
     cmp ecx,ebp
   __DO__	; Do fill it again
     push ebp		; total length to fill
     push ecx		; length already filled
     push ebx		; ebx: buffer address
     call [Next_cluster]; al,cx,dx: next cluster to read.
     pop ebx
     or al,al		; al: number of sectors to read
     jz End_Of_File
     push ax
     call Read_Sectors
     __({"Cluster Read"},CRLF)
     z_z
     pop ax
     pop ecx
     pop ebp
     jc Disk_IO_Error
     clr(edx)
     mov dl,al
     shl edx,#LOG_SECT_SIZE
     add ecx,edx
     add ebx,edx
   __WEND__
 ;; full enough, then give the caller what he asked
   pop ebx ; original buffer
   mov eax,ecx ; length actually read.
   sub eax,ebp
   mov dword [over_fill_len],eax
   zdwoz({"Overfill length now 0x"},eax,{})
   mov eax,ebx
   add eax,ebp
   mov dword [over_fill],eax
   zdwoz({" at 0x"},eax,{CRLF})
   __({"Done Reading once"},CRLF)
   z__z
   ret

 End_Of_File:
   __({"End of file reached"},CRLF)
   z__z
   pop ecx ; total (unsufficient) length of data read
   pop ebp ; length asked
   pop ebx ; old buffer value
   mov eax,ebx
   add eax,ecx
   mov dword [over_fill],eax
   clr(eax)
   mov dword [over_fill_len],eax
   ret

;;; On Errors,
 Disk_IO_Error:
   _PrByte_({CRLF,"Error 0x"},[ErrorCode],{CRLF})
   mov si,#db_zero
   pop ebx ; original ebx
   pop ax ; return address to Main
   br Exit


DoReadOnce:
; calls Read_Once with abortion in case of end-of-file reached.
; See Read_Once just above for specifications.
   call Read_Once
   cmp ecx,ebp
   jb Premature_end_of_file
   ret
 Premature_end_of_file:
   lsa(si,{"Unexpected end of file"},CRLF,0)
   pop ax; return address
   br Exit


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Load Mode 'M' ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

M_Init_disk:
    mov [pCluster],MAP_FILE
    ret


M_Next_cluster:
	push si
	mov si,[pCluster]
	lodsw
	mov cx,ax
	lodsw
	mov dh,ah
	__IF__(nz)
	  or al,al
	__THEN__
	  mov [pCluster],si
	__ENDIF__
	pop si
	mov dl,[BIOS_DL]
	ret
;Entry format:
;BIOS_CX:	.word	0x0002		; CH:cyl&0xFF, CL:sect+((cyl>>8)<<6)
;BIOS_AL:	.byte	0x02		; Int 13h parameters to load the LOADER
;BIOS_DH:	.byte	0x00		; head


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Load Mode 'D' ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

D_Init_disk:
; Unsupported yet
	mov al,#'D
	br Load_Mode_Unsupported
; Look in the root directory, search for the system file, see length and
; cluster.

; ...



D_Next_cluster:

; memorized part of cluster number is next sectors to load.
; while we have a contiguous (without even track limit) zone of sectors on
; disk, look in the FAT for next cluster
; When we reach a limit (or when finished), load the contiguous part.
; memorize next part of cluster to load

; Look in the FAT for the cluster number
; if 12 bit FAT, all the FAT fits in buffer
; if 16 bit FAT, have bits of it; happily a single entry won't be split
; on two sectors.

; ...


;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Load Mode 'H' ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; The LoadParam is used as sect/trk (low) & #cyl (hi).

H_Init_disk:
; Leave track 0 head 0 for the boot sector + fake boot sector + loader
; floppies have less than 255 tracks and 2 heads, (thus) less than 65535 sects.
   mov word [CurrentTrack],#0x0100	; already read Cyl 0 (lo), Hd 1 (hi)
   mov ax,[total_num_sect]
   shr ax,#1				; number of heads=1 ; CF should be 0
   div byte [sect_per_track]
   ; we should have ah=0 (all sectors used in declared surfaces&cylinders)
   mov [Max_Tracks_on],al
   ret


H_Next_cluster:
   mov ax,[LoadArg]	; real max tracks & max sects/tracks
   mov dx,[CurrentTrack]
   __IF__(z)
     xor dh,#1	; change head
   __THEN__
     inc dl	; change track
   __ENDIF__
   cmp dl,[Max_Tracks_on]
   ja  extra_track
   mov cl,byte [sect_per_track]
   sub al,cl
   inc cl
 H_done:
   mov [CurrentTrack],dx
   mov ch,dl
   mov dl,[BIOS_DL]
   ret
 extra_track:
   mov cl,#1
   cmp dl,ah	; only ah real tracks.
   jbe H_done
   clr(al)
   ret



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Load Mode 'C' ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
C_Init_disk:
	__({"'C' Init disk"},CRLF)
	mov ebx,[hidden_sect]
  ; we store the next sector to read in CurrentSector
	movzx eax,byte [LoadArg]
	add eax,ebx
	mov [CurrentSector],eax
  ; we store the maximal value to read in MaxSector
	movzx eax,word [reserved_sect]
	add eax,ebx
	mov [MaxSector],eax
	;z__z
	ret

C_Next_cluster:
	clr(al)
	mov esi,[CurrentSector]
	mov edi,[MaxSector]
	sub edi,esi
	jb CNcret
	push esi
	mov eax,esi
	call SectorNumToBIOS_Args
	mov dl,cl
	and dl,#0x3F			; sector
	clr(eax)
	mov al,[sect_per_track]
	sub al,dl
	inc al
	__IF__(a)
	  cmp eax,edi
	__THEN__
	  mov eax,edi
	  ;STOP(fs,0,79,24,'S,0x0F)
	__ENDIF__
	pop esi
	add esi,eax
	mov [CurrentSector],esi
	mov dl,[BIOS_DL]
  CNcret:
	;__({"'C' Next Cluster Output:"},CRLF)
	;z__z
	ret


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Load Mode 'L' ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
L_Init_disk:
	__({"'L' Init disk"},CRLF)
	; All done in Test_LILO and End_LILOs_work
	ret

L_Next_cluster:
	__({"'L' Next cluster"},CRLF)
	clr(al)		; end_of_file
	ret


;;;;;;;;;;;;;;;;;;;;;;;;;; Low-Level Disk Functions ;;;;;;;;;;;;;;;;;;;;;;;;;;;

; From Linux:
; * This procedure turns off the floppy drive motor, so
; * that we enter the kernel in a known state, and
; * don't have to worry about it later.
kill_motor:
	push dx
	mov dx,#0x3F2
	clr(al)
	out dx,al
	pop dx
	ret

__end__		; undivert et all


;;;;;;;;;;;;;;;;;;;;;;;; Signature for Linuxish loading ;;;;;;;;;;;;;;;;;;;;;;
__dataseg__
	.align 4

setup_sig1:	.word	Lxsig_1
setup_sig2:	.word	Lxsig_2

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; End of LOADER.com ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	.bss
	.align 4
BufferBeg:
